
from colibris import utils

from .base import AuthorizationBackend


class ModelBackend(AuthorizationBackend):
    def __init__(self, model, account_field, **kwargs):
        self.model = utils.import_member(model)
        self.account_field = account_field

        super().__init__(**kwargs)

    def get_account_field(self):
        return getattr(self.model, self.account_field)




from colibris import persist


async def test_health_check_healthy(web_app_client):
    resp = await web_app_client.get('/health')
    assert resp.status == 200

    j = await resp.json()
    assert j == 'healthy'


async def test_health_check_db_down(web_app_client):
    persist.get_database().drop()

    resp = await web_app_client.get('/health')
    assert resp.status == 500

    j = await resp.json()
    assert j['code'] == 'unhealthy'




import os

from colibris.conf import settings
from colibris.authorization import ANY_PERMISSION

from __packagename__ import views
from __packagename__.constants import ROLE_ADMIN, ROLE_REGULAR

#
# Routes example:
#
# ROUTES = [
#     ('GET',     r'/users/me',           views.get_me,           ANY_PERMISSION),
#     (None,      r'/users',              views.UsersView,        {ROLE_ADMIN}),
#     (None,      r'/users/{id:\d+}',     views.UserView,         {ROLE_ADMIN})
# ]
#

#
# Add static routes, for development purposes
#
# if settings.DEBUG:
#     STATIC_ROUTES = [
#         (os.path.join(settings.PROJECT_PACKAGE_DIR, 'static'), '/static'),
#     ]
#





class TaskQueueException(Exception):
    pass


class TimeoutException(TaskQueueException):
    def __init__(self, timeout):
        super().__init__('task did not complete within {} seconds'.format(timeout))


class TaskExecException(TaskQueueException):
    def __init__(self, exc_str):
        self.exc_str = exc_str

    def __str__(self):
        return 'exception while executing task:\n{}'.format(self.exc_str)




import logging
import os
import pytest
import _pytest.config.argparsing

from colibris.conf import settings

from .base import BaseCommand


logger = logging.getLogger(__name__)

_PLUGINS = [
    'aiohttp.pytest_plugin',
    'colibris.test'
]

_TESTS_DIR = 'tests'


class TestCommand(BaseCommand):
    ADD_HELP = False

    def initialize(self):
        # Adjust main program invocation (should be "manage.py test")
        _pytest.config.argparsing.Parser.prog = self.make_prog()

        # Prepare database settings for testing
        if 'name' in settings.DATABASE or 'name' in settings.TEST_DATABASE:
            settings.DATABASE.update(settings.TEST_DATABASE)
            if 'name' not in settings.TEST_DATABASE:
                settings.DATABASE['name'] = 'test_' + settings.DATABASE['name']

            settings.DATABASE['create'] = True  # We want dbs to be created when running tests

        # colibris.setup() will be called at the setup phase of each test

    def parse_arguments(self, parser, args):
        # Use pytest internal arg parser instead of the command's default parser.
        return None

    def execute(self, options):
        has_file_or_dir = any(True for arg in self.args if not arg.startswith('-'))

        # Run tests from the project's tests subpackage unless a file or directory is given as argument
        if not has_file_or_dir:
            tests_dir = os.path.join(settings.PROJECT_PACKAGE_DIR, _TESTS_DIR)
            os.chdir(tests_dir)

        plugins = list(_PLUGINS)
        # Use project's fixtures as a plugin so that project-specific fixtures are loaded
        plugins.append('{}.tests.fixtures'.format(settings.PROJECT_PACKAGE_NAME))

        return pytest.main(self.args, plugins=plugins)





class TemplateException(Exception):
    pass


class TemplateNotFound(TemplateException):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return 'template not found: {}'.format(self.name)


class TemplateSyntaxError(TemplateException):
    def __init__(self, error):
        self.error = error

    def __str__(self):
        return 'template syntax error: {}'.format(self.error)




from peewee_migrate import Router

import colibris

from colibris import persist

from .fixtures import web_app_client


def pytest_runtest_setup():
    colibris.setup()
    if persist.is_enabled():
        router = Router(persist.get_database(), migrate_dir=persist.get_migrations_dir())
        router.run()


def pytest_runtest_teardown():
    if persist.is_created() and not persist.get_database().is_dropped():
        persist.get_database().drop()




class APIException(Exception):
    pass


class BaseJSONException(APIException):
    def __init__(self, code, message, status, details=None):
        self.code = code
        self.message = message
        self.details = details
        self.status = status

    def __str__(self):
        return self.code


class InvalidRequest(BaseJSONException):
    def __init__(self, code, message):
        super().__init__(code, message, status=400)


class UnauthenticatedException(BaseJSONException):
    def __init__(self):
        super().__init__(code='unauthenticated', message='The request cannot be associated to an account.', status=401)


class ForbiddenException(BaseJSONException):
    def __init__(self):
        super().__init__(code='forbidden', message='Access to requested resource is forbidden.', status=403)


class NotFoundException(BaseJSONException):
    def __init__(self, resource='resource'):
        super().__init__('not_found',
                         'The requested {} cannot be found.'.format(resource),
                         status=404)


class DuplicateModelException(InvalidRequest):
    def __init__(self, model, field):
        super().__init__('duplicate_{}'.format(field),
                         'A {} with this {} already exists.'.format(model._meta.name, field))


class ModelNotFoundException(NotFoundException):
    def __init__(self, model):
        super().__init__(model._meta.name)




from .base import AuthorizationBackend


class NullBackend(AuthorizationBackend):
    def __init__(self, **kwargs):
        pass

    def authorize(self, account, method, path, required_permissions):
        # Consider any authenticated request authorized

        return True




import hashlib
import hmac

from http.cookies import Morsel

from colibris.conf import settings
from colibris.api.exceptions import BaseJSONException

from . import get_account
from .exceptions import AuthenticationException


DEFAULT_VALIDITY_SECONDS = 3600 * 24 * 30
SAFE_METHODS = {'GET', 'HEAD', 'OPTIONS', 'TRACE'}

_REQUEST_CSRF_TOKEN_ITEM_NAME = 'csrf_token'
_SECRET_KEY = settings.SECRET_KEY.encode() if settings.SECRET_KEY else b''


class CookieException(AuthenticationException):
    pass


class CSRFValidationException(BaseJSONException):
    def __init__(self):
        super().__init__(code='forbidden', message='CSRF token is either invalid or missing.', status=403)


class CookieBackendMixin:
    def __init__(self, cookie_name=None, cookie_domain=None, validity_seconds=DEFAULT_VALIDITY_SECONDS,
                 csrf_token_cookie_name=None, csrf_token_header_name=None, **kwargs):

        self.cookie_name = cookie_name
        self.cookie_domain = cookie_domain
        self.validity_seconds = validity_seconds
        self.csrf_token_cookie_name = csrf_token_cookie_name
        self.csrf_token_header_name = csrf_token_header_name

    def is_csrf_enabled(self):
        return bool(self.csrf_token_cookie_name and self.csrf_token_header_name)

    def cookie_login(self, response, persistent, token):
        if not self.cookie_name:
            raise CookieException('login attempt without a configured cookie name')

        cookie = Morsel()
        cookie.set(self.cookie_name, token, token)
        cookie['httponly'] = True
        cookie['secure'] = not settings.DEBUG
        cookie['path'] = '/'
        if self.cookie_domain:
            cookie['domain'] = self.cookie_domain

        if persistent:
            cookie['expires'] = self.validity_seconds

        response.cookies[self.cookie_name] = cookie

    def cookie_logout(self, response):
        if not self.cookie_name:
            raise CookieException('logout attempt without a configured cookie name')

        self.clear_cookie(response, self.cookie_name)
        if self.csrf_token_cookie_name:
            self.clear_cookie(response, self.csrf_token_cookie_name)

    def clear_cookie(self, response, name):
        cookie = Morsel()
        cookie.set(name, 'invalid', 'invalid')
        cookie['httponly'] = True
        cookie['secure'] = not settings.DEBUG
        cookie['path'] = '/'
        cookie['expires'] = 0
        if self.cookie_domain:
            cookie['domain'] = self.cookie_domain

        response.cookies[name] = cookie

    def verify_csrf(self, request, account):
        # Skip verification for safe methods
        if request.method in SAFE_METHODS:
            return

        try:
            incoming_csrf_token = request.headers[self.csrf_token_header_name]

        except KeyError:
            raise CSRFValidationException()

        expected_csrf_token = self.get_csrf_token(request, account)

        if incoming_csrf_token != expected_csrf_token:
            raise CSRFValidationException()

    def set_csrf(self, request, response):
        token = self.get_csrf_token(request)
        if not token:
            return

        cookie = Morsel()
        cookie.set(self.csrf_token_cookie_name, token, token)
        cookie['secure'] = not settings.DEBUG
        cookie['path'] = '/'
        if self.cookie_domain:
            cookie['domain'] = self.cookie_domain

        response.cookies['csrf_token'] = cookie

    @staticmethod
    def get_csrf_token(request, account=None):
        token = request.get(_REQUEST_CSRF_TOKEN_ITEM_NAME)
        if token:
            return token

        account = account or get_account(request)
        if account is None:
            return

        account_id = str(account).encode()
        token = hmac.new(_SECRET_KEY, account_id, hashlib.sha256).hexdigest()

        request[_REQUEST_CSRF_TOKEN_ITEM_NAME] = token

        return token




from aiohttp.web import Response

from colibris import api
from colibris import template


def get_object_or_404(model, pk, select_related=None):
    select_related = set(select_related or ())

    try:
        q = model.select(model, *select_related).where(model._meta.primary_key == pk)
        for m in select_related:
            q = q.join(m)

        return q.get()

    except model.DoesNotExist:
        raise api.ModelNotFoundException(model)


def html_response(body=None, status=200, reason=None, headers=None, content_type='text/html'):
    return Response(body=body, status=status, reason=reason,
                    headers=headers, content_type=content_type)


def html_response_template(template_name=None, status=200, reason=None, headers=None, content_type='text/html',
                           **context):

    return html_response(body=template.render(template_name, **context),
                         status=status, reason=reason, headers=headers, content_type=content_type)




from aiohttp import web

from colibris import api
from colibris import app
from colibris import authentication
from colibris import authorization


@web.middleware
async def handle_auth(request, handler):
    if request.match_info.http_exception is not None:
        raise request.match_info.http_exception

    authorization_info = app.route_auth_mapping.get(request.match_info.route)
    method = request.method
    path = request.match_info.route.resource.canonical

    request = authentication.process_request(request)

    # Only go through authentication if route specifies permissions;
    # otherwise route is considered public.
    if authorization_info is not None:
        try:
            account = authentication.authenticate(request)

        except authentication.AuthenticationException:
            raise api.UnauthenticatedException()

        if not authorization.authorize(account, method, path, authorization_info):
            raise api.ForbiddenException()

    response = await handler(request)
    response = authentication.process_response(request, response)

    return response




import code

from colibris import utils
from colibris.conf import settings

from .base import BaseCommand


class ShellCommand(BaseCommand):
    def execute(self, options):
        # have models automatically imported into locals
        loc = locals()
        models = utils.import_module_or_none('{}.models'.format(settings.PROJECT_PACKAGE_NAME))
        if models:
            loc['models'] = models

        code.interact(local=loc)




import pytest

from __packagename__ import models

#
# TEST_USER_JWT = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.' \
#                 'eyJzdWIiOiJ0ZXN0X3VzZXIifQ.' \
#                 'YjAP-wlFcAirvdSYEphXXPh3XJ6jfbWOqi44SXqVYrw'
#
#
# @pytest.fixture
# def test_user():
#     yield models.User.create(username='test_user', password='test_password',
#                              first_name='Test', last_name='User',
#                              email='testuser@example.com')
#
#
# @pytest.fixture
# def write_right(test_user):
#     yield models.Right.create(user=test_user, resource='*', operations='w')
#




import logging

from colibris.conf import settings

from .base import AuthenticationBackend
from .exceptions import *


_REQUEST_ACCOUNT_ITEM_NAME = 'account'


logger = logging.getLogger(__name__)


def authenticate(request):
    account = AuthenticationBackend.get_instance().authenticate(request)
    request[_REQUEST_ACCOUNT_ITEM_NAME] = account

    return account


def get_account(request):
    return request.get(_REQUEST_ACCOUNT_ITEM_NAME)


def login(request, account, persistent):
    logger.debug('logging in account "%s" (persistent=%s)', account, persistent)

    request[_REQUEST_ACCOUNT_ITEM_NAME] = account
    AuthenticationBackend.get_instance().login(request, persistent)


def logout(request):
    account = request.pop(_REQUEST_ACCOUNT_ITEM_NAME, None)
    if account:
        logger.debug('logging out account "%s"', account)
        AuthenticationBackend.get_instance().logout(request)


def process_request(request):
    return AuthenticationBackend.get_instance().process_request(request)


def process_response(request, response):
    return AuthenticationBackend.get_instance().process_response(request, response)


def setup():
    AuthenticationBackend.configure(settings.AUTHENTICATION)




from colibris import utils

from .base import AuthenticationBackend


class ModelBackend(AuthenticationBackend):
    def __init__(self, model, active_field=None, inactive_field=None, **kwargs):
        self.model = utils.import_member(model)
        self.active_field = active_field
        self.inactive_field = inactive_field

        super().__init__(**kwargs)

    def lookup_account(self, auth_data):
        value = self.get_lookup_value(auth_data)
        field = self.get_lookup_field()
        query = (field == value)

        if self.active_field:
            query = query & (self.get_active_field() == True)  # noqa: E712

        if self.inactive_field:
            query = query & (self.get_inactive_field() == False)  # noqa: E712

        try:
            return self.model.select().where(query).get()

        except self.model.DoesNotExist:
            return None

    def get_lookup_field(self):
        raise NotImplementedError

    def get_lookup_value(self, auth_data):
        raise NotImplementedError

    def get_active_field(self):
        return getattr(self.model, self.active_field)

    def get_inactive_field(self):
        return getattr(self.model, self.inactive_field)




import logging

from aiohttp import web
from aiohttp_apispec import validation_middleware
from webargs import aiohttpparser


logger = logging.getLogger(__name__)


class HTTPSchemaValidationError(web.HTTPUnprocessableEntity):
    def __init__(self, schema_error, **kwargs):
        super().__init__(**kwargs)

        self.details = schema_error.messages


@web.middleware
async def handle_schema_validation(request, handler):
    return await validation_middleware(request, handler)


@aiohttpparser.parser.error_handler
def _handle_schema_validation_error(error, req, schema, status_code, headers):
    raise HTTPSchemaValidationError(error)




import peewee


_database_proxy = peewee.DatabaseProxy()


class Model(peewee.Model):
    # This is currently necessary for aiohttp-apispec request data validation
    def __iter__(self):
        return ((k, v) for (k, v) in self.__data__.items())

    def update_fields(self, fields):
        for n, v in fields.items():
            setattr(self, n, v)

    class Meta:
        database = _database_proxy


def set_database(database):
    _database_proxy.initialize(database)





from aiohttp import web
from aiohttp_apispec import docs, request_schema, response_schema

from colibris import api
from colibris import authentication
from colibris.schemas import many_envelope
from colibris.shortcuts import get_object_or_404

from __packagename__ import models
from __packagename__ import schemas

#
# View examples:
#
#
# @docs(tags=['Users'],
#       summary='Reveal details about the current user')
# @response_schema(schemas.UserSchema())
# async def get_me(request):
#     user = authentication.get_account(request)
#     result = schemas.UserSchema().dump(user)
#
#     return web.json_response(result)
#
#
# class UsersView(web.View):
#     @docs(tags=['Users'],
#           summary='List all users')
#     @response_schema(many_envelope(schemas.UserSchema))
#     async def get(self):
#         users = models.User.select().order_by(models.User.username.asc())
#         result = schemas.UserSchema(many=True).dump(list(users))
#
#         return web.json_response(result)
#
#     @docs(tags=['Users'],
#           summary='Add a new user')
#     @request_schema(schemas.UserSchema())
#     @response_schema(schemas.UserSchema())
#     async def post(self):
#         data = self.request['data']
#
#         if models.User.select().where(models.User.username == data['username']).exists():
#             raise api.DuplicateModelException(models.User, 'username')
#
#         user = models.User.create(**data)
#         result = schemas.UserSchema().dump(user)
#
#         return web.json_response(result, status=201)
#
#
# class UserView(web.View):
#     @docs(tags=['Users'],
#           summary='Reveal details about a specific user')
#     @response_schema(schemas.UserSchema())
#     async def get(self):
#         user_id = self.request.match_info['id']
#         user = get_object_or_404(models.User, user_id)
#
#         result = schemas.UserSchema().dump(user)
#
#         return web.json_response(result)
#
#     @docs(tags=['Users'],
#           summary='Update an existing user')
#     @request_schema(schemas.UserSchema(partial=True))
#     @response_schema(schemas.UserSchema(partial=True))
#     async def patch(self):
#         user_id = self.request.match_info['id']
#         user = get_object_or_404(models.User, user_id)
#         data = self.request['data']
#
#         if 'username' in data:
#             query = (models.User.username == data['username']) & (models.User.id != user_id)
#             if models.User.select().where(query).exists():
#                 raise api.DuplicateModelException(models.User, 'username')
#
#         user.update_fields(data)
#         user.save()
#
#         result = schemas.UserSchema().dump(user)
#
#         return web.json_response(result)
#
#     @docs(tags=['Users'],
#           summary='Delete a user')
#     async def delete(self):
#         user_id = self.request.match_info['id']
#         if models.User.delete().where(models.User.id == user_id).execute() == 0:
#             raise api.ModelNotFoundException(models.User)
#
#         return web.json_response(status=204)
#




from colibris.conf.schemas import fields, SettingsSchema

#
# class GeneralSettingsSchema(SettingsSchema):
#     DEBUG = fields.Boolean()
#     LISTEN = fields.String()
#     PORT = fields.Integer()
#     SECRET_KEY = fields.String()
#
#    class Meta:
#        sensible_fields = ['SECRET_KEY']
#
#
# class DatabaseSettingsSchema(SettingsSchema):
#     NAME = fields.String()
#     HOST = fields.String()
#     PORT = fields.Integer()
#     USERNAME = fields.String()
#     PASSWORD = fields.String()
#
#     class Meta:
#         prefix = 'DATABASE_'
#         sensible_field = ['PASSWORD']
#

DEBUG = True

LISTEN = '0.0.0.0'
PORT = 8888

SECRET_KEY = 'replace-me-with-random-ascii-string-or-supply-via-environment'

# DATABASE = {
#     'backend': 'colibris.persist.PostgreSQLBackend',
#     'name': '__packagename__',
#     'host': '127.0.0.1',
#     'port': 5432,
#     'username': 'username',
#     'password': 'password'
# }

# AUTHENTICATION = {
#     'backend': 'colibris.authentication.jwt.JWTBackend',
#     'model': '__packagename__.models.User',
#     'identity_claim': 'sub',
#     'identity_field': 'username',
#     'secret_field': 'password'
# }

# AUTHORIZATION = {
#     'backend': 'colibris.authorization.role.RoleBackend',
#     'role_field': 'role'
# }

#
# GeneralSettingsSchema().load_from_env(globals())
# DatabaseSettingsSchema().load_from_env(globals())
#




import importlib

from peewee_migrate import Router

from colibris import persist
from colibris.conf import settings

from .base import BaseCommand


class MakeMigrationsCommand(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('name', help='An optional migration name', type=str, default='auto', nargs='?')

    def execute(self, options):
        router = Router(persist.get_database(), migrate_dir=persist.get_migrations_dir())
        router.create(options.name, auto=importlib.import_module(settings.PROJECT_PACKAGE_NAME))




from . import fixtures

#
# async def test_unauthenticated(web_app_client):
#     resp = await web_app_client.get('/api/users')
#     assert resp.status == 401
#
#     j = await resp.json()
#     assert j['code'] == 'unauthenticated'
#
#
# async def test_authenticated(web_app_client, test_user, write_right):
#     token = fixtures.TEST_USER_JWT
#     resp = await web_app_client.get('/api/users/me', headers={'Authorization': 'Bearer {}'.format(token)})
#     assert resp.status == 200
#




import logging

from colibris import utils

from . import ImproperlyConfigured


logger = logging.getLogger(__name__)


class BackendMixin:
    _instance = None
    _class = None
    _settings = None

    @classmethod
    def configure(cls, settings):
        cls._settings = settings
        cls._instance = None

        try:
            backend_path = settings.pop('backend')

        except KeyError:
            return  # Backend class not specified

        cls._class = utils.import_member(backend_path)
        logger.debug('%s: using class %s', cls.__name__, backend_path)

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            if cls._settings is None:
                raise ImproperlyConfigured('{} has not been configured'.format(cls.__name__))

            if cls._class is None:
                raise ImproperlyConfigured('class for {} has not been specified'.format(cls.__name__))

            cls._instance = cls._class(**cls._settings)
            cls._instance.on_create()

        return cls._instance

    @classmethod
    def is_configured(cls):
        return cls._settings is not None

    @classmethod
    def is_enabled(cls):
        return cls._class is not None

    @classmethod
    def is_created(cls):
        return cls._instance is not None

    def on_create(self):
        pass




def assert_is_envelope(value, count=None, pages=None, page=None, page_size=None):
    assert isinstance(value, dict)

    assert 'count' in value
    assert 'pages' in value
    assert 'page' in value
    assert 'page_size' in value

    if count is not None:
        assert value['count'] == count
    if pages is not None:
        assert value['pages'] == pages
    if page is not None:
        assert value['page'] == page
    if page_size is not None:
        assert value['page_size'] == page_size




from aiohttp import web
from aiohttp_apispec import docs

from colibris import app
from colibris.conf import settings


async def home(request):
    raise web.HTTPFound(settings.API_DOCS_PATH)


@docs(tags=['Service'],
      summary='The health-check endpoint')
async def health(request):
    h = await app.get_health()
    return web.json_response(h)




from colibris.schemas import ModelSchema
from colibris.schemas import fields, validate, pre_load

from __packagename__ import models

#
# Schema example:
#
#
# class UserSchema(ModelSchema):
#     email = fields.String(validate=[validate.Email(error='Invalid email address.'),
#                                     validate.Length(max=128)])
#
#     @pre_load
#     def process_input(self, data, **kwargs):
#         data['email'] = data['email'].lower().strip()
#
#         return data
#
#     class Meta:
#         model = models.User
#         name = 'user'
#         name_plural = 'users'
#




from . import ANY_PERMISSION
from .base import AuthorizationBackend


class RoleBackend(AuthorizationBackend):
    def __init__(self, role_field, **kwargs):
        self.role_field = role_field

        super().__init__(**kwargs)

    def get_role(self, account):
        return getattr(account, self.role_field)

    def authorize(self, account, method, path, required_permissions):
        if required_permissions == ANY_PERMISSION:
            return True  # anyone authenticated is authorized

        return self.get_role(account) in required_permissions




from colibris import persist

#
# Model example:
#
# class User(persist.Model):
#     id = persist.AutoField()
#     username = persist.CharField(max_length=128, index=True, unique=True)
#     password = persist.CharField(max_length=128)
#     first_name = persist.CharField(max_length=64)
#     last_name = persist.CharField(max_length=64)
#     email = persist.CharField(max_length=128, null=True)
#

#
# Another model example:
#
# class Right(persist.Model):
#     id = persist.AutoField()
#     user = persist.ForeignKeyField(User)
#     resource = persist.CharField(max_length=128)
#     operations = persist.CharField(max_length=16)
#




import logging

from colibris.conf import settings

from .base import AuthorizationBackend


ANY_PERMISSION = '*'

logger = logging.getLogger(__name__)


def authorize(account, method, path, authorization_info):
    return AuthorizationBackend.get_instance().authorize(account, method, path, authorization_info)


def setup():
    AuthorizationBackend.configure(settings.AUTHORIZATION)
